// ZX81 version of '3D Death Chase'
// compile with: (using z88dv compiler:)
// zcc +zx81 -lm -create-app 3d_forest_race.c -o 3d_forest_race.bin

#include <input.h>
#include <zx81.h>

//---------------------------------------------------------------
int __FASTCALL__ shake_asm()
{
	#asm
	ld c,200
.shakeloop
	ld hl,zbuff
	ld b,31
.leftloop
        ld a,(hl)
	inc hl
	ld (hl),a
	djnz leftloop
	ld hl,0
	ld de,1000
.delay1 
        add hl,de
        jr nc,delay1
	ld hl,zbuff
	ld de,31
	add hl,de
	ld b,31
.rightloop	
	ld a,(hl)
	dec hl
	ld (hl),a
	djnz rightloop
	ld hl,0
	ld de,1000
.delay2  
        add hl,de
        jr nc,delay2
	dec c
        jr nz, shakeloop	
	#endasm
}
		   
//--------------------------------------------------------------------------
int __FASTCALL__ init_screen()
{
	#asm
	
	ld hl,(16396)  // the D_FILE screen-memory location flag
	inc hl
	ld (screen),hl
	push hl
	ld a,0
	ld b, 23

.loop_init
	push bc
	ld b, 32

.loop_row1
	ld (hl), 0
	inc hl
	djnz loop_row1

	ld (hl), 0x76
	inc hl
	pop bc

	djnz loop_init
	pop hl  // get the start of screen buffer for reurn address
	#endasm
}

//--------------------------------------------------------------------------
// screen draw routine
// hl = screen buffer location. adds 33 to get to next col down
// de = pointer to tree bytes to write in a column
// bc = 33byte adder to hl
// bc (on stack)  = pointer to Z buffer, low-byte goes from 0 to 32 columns
int __FASTCALL__ draw_screen_asm()
{
	#asm	
	ld hl,(screen)
	ld bc,zbuff
// first, write 14 columns at 22 bytes deep......	
.colwrite1
        push bc
        // first, calculate the tree column to draw...
	ex de,hl
	ld hl,treedata  // l is always 0, alligned data
	srl h    // divide hl by 2
	ld a,(bc) // read the z-buffer
	ld l,a
	add hl,hl  // shift hl left 1 bit
	ex de,hl
        // initially, bc is the zbuffer pointer

        ld bc,33	 // length of line to skip over 
 // the main draw core, hl=screen, de=treebytes, bc=adder       
	ld a,(de)  // 1   1 byte
	ld (hl), a //     1 byte
	add hl,bc  //     1 byte
	inc de     //     1 byte
	ld a,(de)  // 2
	ld (hl), a
	add hl,bc
	inc de
	ld a,(de)  // 3
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 4
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 5
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 6
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 7
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 8
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 9
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 10
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 11
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 12
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 13
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 14
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 15
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 16
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 17
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 18
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 19
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 20
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 21
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 22
	ld (hl), a	
	ld bc,692   // so 33x21-1 bytes to rewind hl for next col.
	sub hl,bc	
	pop bc
        inc c
	ld a,14  // 14 columns to write at 22-deep....
	sub c
	jr nz, colwrite1
	
// now, write 4 columns at 20-bytes deep..........	
.colwrite2
        push bc
        // first, calculate the tree column to draw...
	ex de,hl
	ld hl,treedata  // l is always 0, alligned data
	srl h    // divide hl by 2
	ld a,(bc) // read the z-buffer
	ld l,a
	add hl,hl  // shift hl left 1 bit
	ex de,hl
        // initially, bc is the zbuffer pointer

        ld bc,33	 // length of line to skip over 
 // the main draw core, hl=screen, de=treebytes, bc=adder       
	ld a,(de)  // 1   1 byte
	ld (hl), a //     1 byte
	add hl,bc  //     1 byte
	inc de     //     1 byte
	ld a,(de)  // 2
	ld (hl), a
	add hl,bc
	inc de
	ld a,(de)  // 3
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 4
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 5
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 6
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 7
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 8
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 9
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 10
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 11
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 12
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 13
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 14
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 15
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 16
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 17
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 18
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 19
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 20
	ld (hl), a  	
	ld bc,626   // so 33x19-1 bytes to rewind hl for next col.
	sub hl,bc	
	pop bc
        inc c
	ld a,18  /// loop from 14 to 18 with 20 columns 
	sub c
	jr nz, colwrite2

// lastly, write remaining 14 columns at 22 bytes deep.......	
.colwrite3
        push bc
        // first, calculate the tree column to draw...
	ex de,hl
	ld hl,treedata  // l is always 0, alligned data
	srl h    // divide hl by 2
	ld a,(bc) // read the z-buffer
	ld l,a
	add hl,hl  // shift hl left 1 bit
	ex de,hl
        // initially, bc is the zbuffer pointer

        ld bc,33	 // length of line to skip over 
 // the main draw core, hl=screen, de=treebytes, bc=adder       
	ld a,(de)  // 1   1 byte
	ld (hl), a //     1 byte
	add hl,bc  //     1 byte
	inc de     //     1 byte
	ld a,(de)  // 2
	ld (hl), a
	add hl,bc
	inc de
	ld a,(de)  // 3
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 4
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 5
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 6
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 7
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 8
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 9
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 10
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 11
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 12
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 13
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 14
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 15
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 16
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 17
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 18
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 19
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 20
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 21
	ld (hl), a  
	add hl,bc
	inc de
	ld a,(de)  // 22
	ld (hl), a	
	ld bc,692   // so 33x21-1 bytes to rewind hl for next col.
	sub hl,bc	
	pop bc
        inc c
	ld a,32  // loop from 18 to 32 columns to write at 22-deep....
	sub c
	jr nz, colwrite3

	
	
	ret
	#endasm
}

//--------------------------------------------------------------
// read map and generate 32 zbuffer tree columns to draw later
void generate_zbuff_asm()
{
   #asm 
   ld bc,zbuff
   ld de,map
   ld hl,alignmap
   
   ld l,0  // 16 (x8) columns to process....
.zcol_loop 
   sla l
   sla l
   sla l  // multiply l by 8 initially
   // depth=8
   dec l
   ld e,(hl) 
   ld a,(de)
   and 128 // check for match
   jr nz, endcol
   // depth=7
   dec l
   ld e,(hl)
   ld a,(de)
   and 64  // check for match 
   jr nz, endcol
   // depth=6
   dec l
   ld e,(hl)
   ld a,(de)
   and 32  // check for match
   jr nz, endcol
   // depth=5
   dec l
   ld e,(hl)
   ld a,(de)
   and 16  // check for match
   jr nz, endcol
   // depth=4
   dec l
   ld e,(hl)
   ld a,(de)
   and 8  // check for match
   jr nz, endcol
   // depth=3
   dec l
   ld e,(hl)
   ld a,(de)
   and 4  // check for match
   jr nz, endcol
   // depth=2
   dec l
   ld e,(hl)
   ld a,(de)
   and 2  // check for match
   jr nz, endcol
   // depth=1
   dec l
   ld e,(hl)
   ld a,(de)
   and 1  // check for match at depth 7
   jr nz, endcol   
   // give up checking, just load a blank background.
   ld a,0  // use xor a instead? no big deal
   jr endcol_write
   
.endcol
   inc  h  // switch to reading tree-columns
   ld a,(hl)  // read the correct tree column
   dec h  // switch back
.endcol_write
   ld (bc),a  // write the approprate tree column  
   inc bc 
   sra l
   sra l
   sra l   // divide l by 8 now, bit 7 unaffected
   jr nz, zcol_loop
   #endasm   
}

//-------------------------------------------------
int __FASTCALL__ get_zbuff()
{
  #asm
  ld hl, zbuff
  #endasm
}

//-------------------------------------------------
int __FASTCALL__ get_map()
{
  #asm
  ld hl, map
  #endasm
}

//---------------------------------------------------
void shake(uchar *zbuff)
{
  int i;
    for (i=0; i<31; i++) zbuff[i]=zbuff[i+1];
    draw_screen_asm();
    for (i=31; i>0; i--) zbuff[i]=zbuff[i-1];
    draw_screen_asm();
    for (i=0; i<31; i++) zbuff[i]=zbuff[i+1];
    draw_screen_asm();
    for (i=31; i>0; i--) zbuff[i]=zbuff[i-1];
    draw_screen_asm();
}

//--------------------------------------------------------
void left()
{
   #asm
   ld hl,map
   ld b,40
.mapleftloop
   inc hl
   ld a,(hl)
   dec hl
   ld (hl),a  
   inc hl
   djnz mapleftloop
   #endasm
}

//---------------------------------------------------------
void right()
{
   // for (i=40; i>0; i--) {map[i]=map[i-1]; map[0]=0;}
   #asm
   ld hl,map
   ld l,40
.maprightloop
   dec l
   ld a,(hl)
   inc l
   ld (hl),a
   dec l
   jr nz, maprightloop
   #endasm
}

//---------------------------------------------------------
void write_score_lives()
{
  #asm
  ld hl,screen
  ld de,774
  add hl,de
  ld de,score
  ld b,5
.scoreloop1
  ld a,(de)
  ld (hl),a
  inc hl
  inc de
  djnz scoreloop1
  
  ld de,16
  add hl,de
  ld de,lives
  ld b,5
.scoreloop2
  ld a,(de)
  ld (hl),a
  inc hl
  inc de
  djnz scoreloop2 
   
  ld de,7
  add hl,de
  ld de,zeros
  ld b,5
.scoreloop3
  ld a,(de)
  ld (hl),a
  inc hl
  inc de
  djnz scoreloop3
  ld de,18
  add hl,de
  ld a,0xa1
  ld (hl),a
  #endasm
}

//-----------------------------------------------------
void increase_score()
{
  #asm
  ld c,0xa6
  ld hl,screen
  ld de,811
  add hl,de
  
  // digit 5
  ld a,(hl)
  inc a
  ld (hl),a
  xor 0xa6
  ret nz
  ld a,0x9c
  ld (hl),a
  dec hl
  
  // digit 4
  ld a,(hl)
  inc a
  ld (hl),a
  xor 0xa6
  ret nz  
  ld a,0x9c
  ld (hl),a
  dec hl
  
  // digit 3
  ld a,(hl)
  inc a
  ld (hl),a
  xor 0xa6
  ret nz
  ld a,0x9c
  ld (hl),a
  dec hl
  
  // digit 2
  ld a,(hl)
  inc a
  ld (hl),a
  xor 0xa6
  ret nz  
  ld a,0x9c
  ld (hl),a
  dec hl
  
  // digit 1
  ld a,(hl)
  inc a
  ld (hl),a
  xor 0xa6
  ret nz  
  ld a,0x9c
  ld (hl),a
  ret
  #endasm
}

//------------------------------------------------------
void game_over()
{
  #asm
  ld hl,screen
  ld de,353
  add hl,de
  ld de,gameover
  ld b,9
.gameoverloop
  ld a,(de)
  ld (hl),a
  inc hl
  inc de
  push bc
  push hl
  ld hl,0
  ld bc,10
gwait:
  add hl,bc
  jr nc,gwait
  pop hl
  pop bc
  djnz gameoverloop  
  #endasm
}
*/

//-------------------------------------------------------
int main()
{
   uchar *screen;
   char score_str[16]={ 0x38, 0x28, 0x34, 0x37, 0x2a, 0,0,0,0,0,0,0,0,0,  0x99 };
   int score;
   uchar *map;  // 40x8 trees map
   uchar *zbuff; 
   int i,n;
   char lives;
   int difficulty=1500;  // useful between 500 (easy) and 1500 (hard)
 
   map=(uchar *)get_map();
   zbuff=(uchar *)get_zbuff();
start:   
   screen=init_screen();
   for (i=0; i<32; i++) screen[i+726]=128;
   for (i=0; i<32; i++) screen[i+759]=128;
   screen[737]=0; screen[738]=0; screen[745]=0; screen[746]=0;
   screen[739]=129; screen[744]=130;
   //screen[772]=133; screen[777]=5;
   screen[772]=128; screen[777]=128;
   screen[770]=0; screen[771]=0; screen[778]=0; screen[779]=0;
   for (n=660; n<792; n+=33) {
     for (i=14; i<18; i++) screen[i+n]=128;
   }
   // initial straight handlebars.....
   screen[674]=138; screen[677]=138;
   screen[707]=135;   screen[710]=4;
   // write score 00000 and lives 5......
   write_score_lives();
   for (i=0; i<40; i++) map[i]=0;
   generate_zbuff_asm();
   draw_screen_asm();
   printf("        3d forest race\n");
   printf("\n\n");
   printf("       5=left   8=right\n");
   printf("\n");
   printf("  choose difficulty to start \n");
   printf("  1=easy  2=medium  3=hard ");
   msleep(100);
   difficulty=0;
   while (difficulty==0) {
      i=in_Inkey();
      if (i=='1') difficulty=500;
      if (i=='2') difficulty=800;
      if (i=='3') difficulty=1400;     
   }
   
   draw_screen_asm();
   printf("     ready...");
   msleep(150);
   printf("  steady....");
   msleep(150);
   printf("  go\n");
   msleep(100);
   
   
   while(1) {
     for (i=0; i<40; i++) if (rand()<difficulty) map[i]|=1;
     generate_zbuff_asm();
     draw_screen_asm();
     if (map[20]>=128) {   // crash!
         shake(zbuff);
	 map[20]=0;  // clear path ahead for restart
	 msleep(200);
	 lives=screen[785];
	 lives=lives-1;
	 screen[785]=lives;
	 if (lives==0x9c) {
	   game_over();
	   msleep(500);
	   while (in_Inkey()==0);
	   //goto start;
	   write_score_lives();
	   for (i=0; i<40; i++) map[i]=0;
	 }
     }
     for (i=0; i<40; i++) map[i]<<=1;
     i=in_Inkey();
     if (i=='5') {  // turn left....
        //for (i=0; i<40; i++) map[i]=map[i+1];
	left();
        screen[674]=8; screen[677]=137;
        screen[707]=137;   screen[710]=4;        
     }
     else if (i=='8') { // turn right....
        //for (i=40; i>0; i--) {map[i]=map[i-1]; map[0]=0;}
	right();
        screen[674]=137; screen[677]=8;
        screen[707]=135;   screen[710]=137;
     }
     else {    // straight ahead
     
        screen[674]=138; screen[677]=138;
        screen[707]=135;   screen[710]=4;
     }
     increase_score();
     //msleep(2);   
  }  
}

//-----------------------------------------------------------------
// 256-byte alligned section for data maps
#asm
section bss_align_512
align 512

// tree widths are: 1,1,1,1,1,2,2,4 = 13 total, plus blank=14

// which map items correspond to a tree at each 8x32 posn
// map 0 us reserved/ empty
.alignmap  
defb \\
07,08,00,10,00,12,15,16, \\
08,00,09,00,13,13,15,16, \\
09,09,10,11,00,13,00,17, \\
10,10,11,12,14,14,16,17, \\
00,11,00,00,00,14,16,17, \\
11,12,12,13,15,15,00,17, \\
12,00,13,14,00,15,17,18, \\
13,13,14,00,16,16,17,18, \\
14,14,00,15,00,16,00,18, \\
15,15,15,16,17,17,18,18, \\
00,16,16,00,00,17,18,19, \\
16,00,17,17,18,18,00,19, \\
17,17,00,18,00,18,19,19, \\
18,18,18,00,19,19,19,19, \\
19,19,19,19,00,19,00,20, \\
20,20,20,20,20,20,20,20, \\
21,21,21,21,00,20,20,20, \\
22,22,22,00,21,21,00,20, \\
23,23,00,22,00,21,21,21, \\
24,00,23,23,22,22,21,21, \\
00,24,24,00,00,22,00,21, \\
25,25,25,24,23,23,22,21, \\
26,26,00,25,00,23,22,22, \\
27,27,26,00,24,24,00,22, \\
28,00,27,26,00,24,23,22, \\
29,28,28,27,25,25,23,22, \\
00,29,00,00,00,25,00,23, \\
30,30,29,28,26,26,24,23, \\
31,31,30,29,00,26,24,23, \\
32,00,31,00,27,27,00,23, \\
33,32,00,30,00,27,25,24, \\
34,33,32,31,28,28,25,24

.aligntree // which tree columns correspond to a tree at each 8x16 posn
defb \\
143,132,000,110,000,088,055,033, \\
143,000,121,000,099,077,066,044, \\
143,132,121,110,000,088,000,011, \\
143,132,121,110,099,077,055,022, \\
000,132,000,000,000,088,066,033, \\
143,132,121,110,099,077,000,044, \\
143,000,121,110,000,088,055,011, \\
143,132,121,000,099,077,066,022, \\
143,132,000,110,000,088,000,033, \\
143,132,121,110,099,077,055,044, \\
000,132,121,000,000,088,066,011, \\
143,000,121,110,099,077,000,022, \\
143,132,000,110,000,088,055,033, \\
143,132,121,000,099,077,066,044, \\
143,132,121,110,000,088,000,011, \\
143,132,121,110,099,077,055,022, \\
143,132,121,110,000,088,066,033, \\
143,132,121,000,099,077,000,044, \\
143,132,000,110,000,088,055,011, \\
143,000,121,110,099,077,066,022, \\
000,132,121,000,000,088,000,033, \\
143,132,121,110,099,077,055,044, \\
143,132,000,110,000,088,066,011, \\
143,132,121,000,099,077,000,022, \\
143,000,121,110,000,088,055,033, \\
143,132,121,110,099,077,066,044, \\
000,132,000,000,000,088,000,011, \\
143,132,121,110,099,077,055,022, \\
143,132,121,110,000,088,066,033, \\
143,000,121,000,099,077,000,044, \\
143,132,000,110,000,088,055,011, \\
143,132,121,110,099,077,066,022

//-----------------------------------------------------------
// the 14 tree graphics, each column 22 bytes.
.treedata  
defb \\
// background
0,0,0,0,0,0,0,0,0,0,0,8,8,8,8,8,8,8,8,8,8,8, \\

// frame 8 (11 to 44)
128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128, \\
128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128, \\
128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128, \\
128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128, \\

// frame 7 (55 to 66)
6,6,6,6,6,128,128,128,128,128,128,128,128,128,128,128,8,8,8,8,8,8, \\
6,6,6,6,6,128,128,128,128,128,128,128,128,128,128,128,8,8,8,8,8,8, \\

// frame 6 (77 to 88)
0,0,6,6,6,6,6,128,128,128,128,128,128,128,8,8,8,8,8,8,8,8, \\
0,0,6,6,6,6,6,128,128,128,128,128,128,128,8,8,8,8,8,8,8,8, \\

// frame 5 (99)
0,0,0,0,0,6,6,6,6,128,128,128,128,8,8,8,8,8,8,8,8,8, \\

// frame 4 (110)
0,0,0,0,0,0,6,6,6,128,128,128,8,8,8,8,8,8,8,8,8,8, \\

// frame 3 (121)
0,0,0,0,0,0,0,0,6,5,5,8,8,8,8,8,8,8,8,8,8,8, \\

// frame 2 (132)
0,0,0,0,0,0,0,0,0,0,5,8,8,8,8,8,8,8,8,8,8,8, \\

// frame 1 (143)
0,0,0,0,0,0,0,0,0,0,4,8,8,8,8,8,8,8,8,8,8,8
defs 204  // pad to 512 bytes

.map   // the 40-byte map, (padded to 256 bytes)
defs 256

.zbuff
defs 32   // alligned 32 8-bit ints for zbuff pointers

.screen
defs 2
.tree8div2
defs 2
.score
defb 0xb8,0xa8,0xb4,0xb7,0xaa
.lives
defb 0xb1, 0xae, 0xbb, 0xaa, 0xb8
.zeros
defb 0x9c,0x9c,0x9c,0x9c,0x9c
.gameover
defb 0x2c,0x26,0x32,0x2a,0x00,0x34,0x3b,0x2a,0x37
#endasm

